import 'dart:ui';

import 'package:flutter/foundation.dart';
import 'package:flutterx_utils/src/extensions/code.dart';
import 'package:flutterx_utils/src/utils/cycle.dart';

/// JsonPath represents a path of keys inside a json like collection
/// The path is composed only by nested [Map] and [List] objects except the last segment that can have any type of value
/// Each path entry has a [parent] which is the parent entry or [root] for the first segment
/// And a [value] which is the key of the current path entry (any type for maps and int index for lists)
/// The [root] element is is the element that does not have a [parent]
@immutable
class JsonPath implements Comparable<JsonPath> {
  static const JsonPath root = JsonPath._(null, null);

  /// Parent element of this entry or null if this is the [root]
  final JsonPath? parent;

  /// Key of the current map/list
  final dynamic value;

  /// Return true if this is the [root] ([parent] is null)
  bool get isRoot => parent == null;

  /// All segments of this path from the last one to [root] (reverse of [segments])
  Iterable<JsonPath> get segmentsToRoot => Cycle<JsonPath>(this, cycle: (current) => current.parent);

  /// All segments of this path from the [root] to this one (reverse of [segmentsToRoot])
  Iterable<JsonPath> get segments => segmentsToRoot.toList(growable: false).reversed;

  const JsonPath._(this.parent, this.value);

  /// Create new entry from [root] representing /[value]
  factory JsonPath(dynamic value) => root.append(value);

  /// Create a path from iterable of keys e.g. 'path'/'to'/'my'/1/'object'
  factory JsonPath.from(Iterable values) {
    var result = JsonPath.root;
    for (final value in values) result = result.append(value);
    return result;
  }

  /// Create a sub-path of this entry having this as a parent and [value] as key
  JsonPath append(dynamic value) => JsonPath._(this, value);

  /// Check weather this [other] is a sub-path of this
  bool containsPath(JsonPath other) => segmentsToRoot.contains(other);

  /// Check if [value] key exists in [target] at parent path element
  bool contains(Object target) {
    final scope = _scope(target);
    if (scope is Map) {
      return scope.containsKey(value);
    } else if (scope is List) {
      return value >= 0 && value < scope.length;
    } else {
      _error('query', scope);
    }
  }

  /// Get element of [target] at this path
  E get<E>(Object target) {
    final scope = _scope(target);
    if (scope is Map) {
      return scope[value];
    } else if (scope is List) {
      return scope[value];
    } else {
      _error('get', scope);
    }
  }

  /// Remove element of [target] at this path
  E remove<E>(Object target) {
    final scope = _scope(target);
    if (scope is Map) {
      return scope.remove(value);
    } else if (scope is List) {
      return scope.removeAt(value);
    } else {
      _error('remove', scope);
    }
  }

  /// Set element of [target] at this path to [data]
  void set(Object target, dynamic data) {
    final scope = _scope(target);
    if (scope is Map) {
      scope[value] = data;
    } else if (scope is List) {
      scope[value] = data;
    } else {
      _error('set', scope);
    }
  }

  @override
  String toString() => isRoot ? '' : '$parent[$value]';

  @override
  int get hashCode => hashValues(parent, value);

  @override
  bool operator ==(Object other) =>
      identical(other, this) || other is JsonPath && other.parent == parent && other.value == value;

  @override
  int compareTo(JsonPath other) => toString().compareTo(other.toString());

  /// [throws] StateError if parent element(s) does not exist in [target]
  Object _scope(Object target) {
    final parent = requireNotNull(this.parent, message: () => 'cannot query root element');
    if (parent.isRoot) return target;
    return requireNotNull(parent.get(target),
        message: () => '$this: ${parent.value} not found in ${target.runtimeType} $target');
  }

  /// [throws] UnsupportedError if [scope] is not either a [Map] or a [List]
  Never _error(String action, Object scope) =>
      throw UnsupportedError('Cannot perform action \'$action\' on path \'$this\' of ${scope.runtimeType} $scope');
}
